@@ -1,6 +1,6 @@
 /********************************************
 da.c
-copyright 2008-2020,2021, Thomas E. Dickey
+copyright 2008-2021,2023, Thomas E. Dickey
 copyright 1991-1994,1995, Michael D. Brennan
 
 This is a source file for mawk, an implementation of
@@ -11,18 +11,18 @@ the GNU General Public License, version 2, 1991.
 ********************************************/
 
 /*
- * $MawkId: da.c,v 1.30 2021/05/29 00:00:01 tom Exp $
+ * $MawkId: da.c,v 1.46 2023/08/16 23:33:50 tom Exp $
  */
 
 /*  da.c  */
 /*  disassemble code */
 
-#include  "mawk.h"
+#include  <mawk.h>
 
-#include  "code.h"
-#include  "bi_funct.h"
-#include  "repl.h"
-#include  "field.h"
+#include  <code.h>
+#include  <bi_funct.h>
+#include  <repl.h>
+#include  <field.h>
 
 typedef struct fdump {
     struct fdump *link;
@@ -32,7 +32,7 @@ typedef struct fdump {
 static DUMP_FUNCS *fdump_list;	/* linked list of all user functions */
 
 #ifdef LOCAL_REGEXP
-#include "regexp.h"
+#include <regexp.h>
 
 typedef struct regex {
     struct regex *link;
@@ -58,7 +58,6 @@ static const OP_NAME simple_code[] =
     { _STOP,      "stop" },
     { FE_PUSHA,   "fe_pusha" },
     { FE_PUSHI,   "fe_pushi" },
-    { A_LENGTH,   "a_length" },
     { A_TEST,     "a_test" },
     { A_DEL,      "a_del" },
     { DEL_A,      "del_a" },
@@ -123,40 +122,85 @@ static const char *jfmt = "%s%s%03d\n";
    /* format to print jumps */
 static const char *tab2 = "\t\t";
 
+#if OPT_TRACE
+static FBLOCK *
+inst2fblock(INST * p)
+{
+    DUMP_FUNCS *q;
+    FBLOCK *result = NULL;
+    for (q = fdump_list; q != NULL; q = q->link) {
+	if (q->data->code == p) {
+	    result = q->data;
+	    break;
+	}
+    }
+    return result;
+}
+#endif
+
 void
-da_string(FILE *fp, const char *str, size_t len)
+da_string2(FILE *fp, const char *value, size_t length, int delim)
 {
     size_t n;
 
-    fputc('"', fp);
-    for (n = 0; n < len; ++n) {
-	UChar ch = (UChar) str[n];
+    fputc(delim, fp);
+    for (n = 0; n < length; ++n) {
+	UChar ch = (UChar) value[n];
 	switch (ch) {
 	case '\\':
-	    fprintf(fp, "\\\\");
+	    fputc(ch, fp);
 	    break;
-	case '"':
-	    fprintf(fp, "\"");
+	case '\a':		/* alert, ascii 7 */
+	    fputs("\\\\", fp);
+	    break;
+	case '\b':		/* backspace, ascii 8 */
+	    fputs("\\b", fp);
+	    break;
+	case '\t':		/* tab, ascii 9 */
+	    fputs("\\t", fp);
+	    break;
+	case '\n':		/* newline, ascii 10 */
+	    fputs("\\n", fp);
+	    break;
+	case '\v':		/* vertical tab, ascii 11 */
+	    fputs("\\v", fp);
+	    break;
+	case '\f':		/* formfeed, ascii 12 */
+	    fputs("\\f", fp);
+	    break;
+	case '\r':		/* carriage return, ascii 13 */
+	    fputs("\\r", fp);
 	    break;
 	default:
-	    if (ch >= 32 && ch < 127)
-		fprintf(fp, "%c", ch);
+	    if (ch == delim)
+		fprintf(fp, "\\%c", ch);
+	    else if (ch >= 32 && ch < 127)
+		fputc(ch, fp);
 	    else
 		fprintf(fp, "\\%03o", ch);
 	    break;
 	}
     }
-    fputc('"', fp);
+    fputc(delim, fp);
 }
 
+void
+da_string(FILE *fp, const STRING * sparm, int delim)
+{
+    da_string2(fp, sparm->str, sparm->len, delim);
+}
+
+#define NORMAL_FORM "%03ld .\t"
+#define ADJUST_FORM  "# patching %s\n    .\t"
+
 INST *
 da_this(INST * p, INST * start, FILE *fp)
 {
     CELL *cp;
+    const char *op_name = da_op_name(p);
 
     /* print the relative code address (label) */
-    fprintf(fp, "%03ld ", (long) (p - start));
-    fprintf(fp, ".\t");
+    fprintf(fp, NORMAL_FORM, (long) (p - start));
 
     switch ((MAWK_OPCODES) (p++->op)) {
 
@@ -165,64 +209,66 @@ da_this(INST * p, INST * start, FILE *fp)
 	switch (cp->type) {
 	case C_RE:
 	    add_to_regex_list(cp->ptr);
-	    fprintf(fp, "pushc\t%p\t/%s/\n", cp->ptr,
-		    re_uncompile(cp->ptr));
+	    fprintf(fp, "%s\t%p\t", op_name, cp->ptr);
+	    da_string(fp, re_uncompile(cp->ptr), '/');
+	    fputc('\n', fp);
 	    break;
 
 	case C_SPACE:
-	    fprintf(fp, "pushc\tspace split\n");
+	    fprintf(fp, "%s\tspace split\n", op_name);
 	    break;
 
 	case C_SNULL:
-	    fprintf(fp, "pushc\tnull split\n");
+	    fprintf(fp, "%s\tnull split\n", op_name);
 	    break;
+
 	case C_REPL:
-	    fprintf(fp, "pushc\trepl\t%s\n",
-		    repl_uncompile(cp));
+	    fprintf(fp, "%s\trepl\t", op_name);
+	    da_string(fp, repl_uncompile(cp), '"');
+	    fputc('\n', fp);
 	    break;
+
 	case C_REPLV:
-	    fprintf(fp, "pushc\treplv\t%s\n",
-		    repl_uncompile(cp));
+	    fprintf(fp, "%s\treplv\t", op_name);
+	    da_string(fp, repl_uncompile(cp), '"');
+	    fputc('\n', fp);
 	    break;
 
 	default:
-	    fprintf(fp, "pushc\tWEIRD\n");;
+	    fprintf(fp, "%s\tWEIRD\n", op_name);;
 	    break;
 	}
 	break;
 
     case _PUSHD:
-	fprintf(fp, "pushd\t%.6g\n", *(double *) p++->ptr);
+	fprintf(fp, "%s\t%.6g\n", op_name, *(double *) p++->ptr);
 	break;
+
     case _PUSHS:
-	{
-	    STRING *sval = (STRING *) p++->ptr;
-	    fprintf(fp, "pushs\t");
-	    da_string(fp, sval->str, sval->len);
-	    fprintf(fp, "\n");
-	    break;
-	}
+	fprintf(fp, "%s\t", op_name);
+	da_string(fp, (STRING *) p++->ptr, '"');
+	fputc('\n', fp);
+	break;
 
     case _MATCH0:
     case _MATCH1:
 	add_to_regex_list(p->ptr);
-	fprintf(fp, "match%d\t%p\t/%s/\n",
-		p[-1].op == _MATCH1, p->ptr,
-		re_uncompile(p->ptr));
+	fprintf(fp, "%s\t%p\t", op_name, p->ptr);
+	da_string(fp, re_uncompile(p->ptr), '/');
+	fputc('\n', fp);
 	p++;
 	break;
 
     case _PUSHA:
-	fprintf(fp, "pusha\t%s\n",
-		reverse_find(ST_VAR, &p++->ptr));
+	fprintf(fp, "%s\t%s\n", op_name, reverse_find(ST_VAR, &p++->ptr));
 	break;
 
     case _PUSHI:
 	cp = (CELL *) p++->ptr;
 	if (cp == field)
-	    fprintf(fp, "pushi\t$0\n");
+	    fprintf(fp, "%s\t$0\n", op_name);
 	else if (cp == &fs_shadow)
-	    fprintf(fp, "pushi\t@fs_shadow\n");
+	    fprintf(fp, "%s\t@fs_shadow\n", op_name);
 	else {
 	    const char *name;
 	    if (cp > NF && cp <= LAST_PFIELD)
@@ -230,139 +276,143 @@ da_this(INST * p, INST * start, FILE *fp)
 	    else
 		name = reverse_find(ST_VAR, &cp);
 
-	    fprintf(fp, "pushi\t%s\n", name);
+	    fprintf(fp, "%s\t%s\n", op_name, name);
 	}
 	break;
 
     case L_PUSHA:
-	fprintf(fp, "l_pusha\t%d\n", p++->op);
-	break;
-
     case L_PUSHI:
-	fprintf(fp, "l_pushi %d\t# frame-number\n", p++->op);
-	break;
-
     case LAE_PUSHI:
-	fprintf(fp, "lae_pushi\t%d\n", p++->op);
-	break;
-
     case LAE_PUSHA:
-	fprintf(fp, "lae_pusha\t%d\n", p++->op);
-	break;
-
     case LA_PUSHA:
-	fprintf(fp, "la_pusha\t%d\n", p++->op);
+	fprintf(fp, "%s\t%d\n", op_name, p++->op);
 	break;
 
     case F_PUSHA:
 	cp = (CELL *) p++->ptr;
 	if (cp >= NF && cp <= LAST_PFIELD)
-	    fprintf(fp, "f_pusha\t%s\n",
-		    reverse_find(ST_FIELD, &cp));
+	    fprintf(fp, "%s\t%s\n", op_name, reverse_find(ST_FIELD, &cp));
 	else
-	    fprintf(fp, "f_pusha\t$%d\n",
-		    field_addr_to_index(cp));
+	    fprintf(fp, "%s\t$%d\n", op_name, field_addr_to_index(cp));
 	break;
 
     case F_PUSHI:
 	p++;
-	fprintf(fp, "f_pushi\t$%d\n", p++->op);
+	fprintf(fp, "%s\t$%d\n", op_name, p++->op);
 	break;
 
     case AE_PUSHA:
-	fprintf(fp, "ae_pusha\t%s\n",
-		reverse_find(ST_ARRAY, &p++->ptr));
-	break;
-
     case AE_PUSHI:
-	fprintf(fp, "ae_pushi\t%s\n",
-		reverse_find(ST_ARRAY, &p++->ptr));
+    case A_PUSHA:
+	fprintf(fp, "%s\t%s\n", op_name, reverse_find(ST_ARRAY, &p++->ptr));
 	break;
 
-    case A_PUSHA:
-	fprintf(fp, "a_pusha\t%s\n",
-		reverse_find(ST_ARRAY, &p++->ptr));
+    case A_LENGTH:
+	{
+	    SYMTAB *stp = (SYMTAB *) p++->ptr;
+
+	    fprintf(fp, ADJUST_FORM, type_to_str(stp->type));
+	    TRACE((NORMAL_FORM, (long) (p - 2 - start)));
+	    TRACE(("patch/alen %s\n", type_to_str(stp->type)));
+
+	    switch (stp->type) {
+	    case ST_VAR:
+		fprintf(fp, "pushi\t%s\t# defer_alen\n", stp->name);
+		break;
+	    case ST_ARRAY:
+		fprintf(fp, "a_pusha\t%s\t# defer_alen\n", stp->name);
+		p[1].ptr = (PTR) bi_alength;
+		break;
+	    case ST_NONE:
+		fprintf(fp, "pushi\t%s\t# defer_alen\n", "@missing");
+		break;
+	    default:
+		bozo("da A_LENGTH");
+		/* NOTREACHED */
+	    }
+	}
+	break;
+
+    case _LENGTH:
+	{
+	    DEFER_LEN *dl = (DEFER_LEN *) p++->ptr;
+	    FBLOCK *fbp = dl->fbp;
+	    short offset = dl->offset;
+	    int type = fbp->typev[offset];
+
+	    fprintf(fp, ADJUST_FORM, type_to_str(type));
+	    TRACE((NORMAL_FORM, (long) (p - 2 - start)));
+	    TRACE(("patch/len %s\n", type_to_str(type)));
+
+	    switch (type) {
+	    case ST_LOCAL_VAR:
+		fprintf(fp, "l_pushi\t@%hd\t# defer_len\n", offset);
+		break;
+	    case ST_LOCAL_ARRAY:
+		fprintf(fp, "la_pusha\t@%hd\t# defer_len\n", offset);
+		p[1].ptr = (PTR) bi_alength;
+		break;
+	    case ST_LOCAL_NONE:
+		fprintf(fp, "pushi\t%s\t# defer_len\n", "@missing");
+		break;
+	    default:
+		bozo("da _LENGTH");
+		/* NOTREACHED */
+	    }
+	}
 	break;
 
     case _PUSHINT:
-	fprintf(fp, "pushint\t%d\n", p++->op);
+	fprintf(fp, "%s\t%d\n", op_name, p++->op);
 	break;
 
     case _BUILTIN:
-	fprintf(fp, "%s\n",
-		find_bi_name((PF_CP) p++->ptr));
-	break;
-
     case _PRINT:
-	fprintf(fp, "%s\n",
-		(PF_CP) p++->ptr == bi_printf
-		? "printf" : "print");
+	fprintf(fp, "%s\n", op_name);
+	++p;
 	break;
 
     case _JMP:
-	fprintf(fp, jfmt, "jmp", tab2, (p - start) + p->op);
-	p++;
-	break;
-
     case _JNZ:
-	fprintf(fp, jfmt, "jnz", tab2, (p - start) + p->op);
-	p++;
-	break;
-
     case _JZ:
-	fprintf(fp, jfmt, "jz", tab2, (p - start) + p->op);
-	p++;
-	break;
-
     case _LJZ:
-	fprintf(fp, jfmt, "ljz", tab2, (p - start) + p->op);
-	p++;
-	break;
-
     case _LJNZ:
-	fprintf(fp, jfmt, "ljnz", tab2 + 1, (p - start) + p->op);
+	fprintf(fp, jfmt, op_name,
+		(strlen(op_name) < 8) ? tab2 : (tab2 + 1),
+		(p - start) + p->op);
 	p++;
 	break;
 
     case SET_ALOOP:
-	fprintf(fp, "set_al\t%03ld\n", (long) (p + p->op - start));
+	fprintf(fp, "%s\t%03ld\n", op_name, (long) (p + p->op - start));
 	p++;
 	break;
 
     case ALOOP:
-	fprintf(fp, "aloop\t%03ld\n", (long) (p - start + p->op));
+	fprintf(fp, "%s\t%03ld\n", op_name, (long) (p - start + p->op));
 	p++;
 	break;
 
     case A_CAT:
-	fprintf(fp, "a_cat\t%d\n", p++->op);
+	fprintf(fp, "%s\t%d\n", op_name, p++->op);
 	break;
 
     case _CALL:
-	fprintf(fp, "call\t%s\t%d\n",
-		((FBLOCK *) p->ptr)->name, p[1].op);
+	fprintf(fp, "%s\t%s\t%d\n", op_name, ((FBLOCK *) p->ptr)->name, p[1].op);
 	p += 2;
 	break;
 
     case _RANGE:
-	fprintf(fp, "range\t%03ld %03ld %03ld\n",
+	fprintf(fp, "%s\t%03ld %03ld %03ld\n", op_name,
 	/* label for pat2, action, follow */
 		(long) (p - start + p[1].op),
 		(long) (p - start + p[2].op),
 		(long) (p - start + p[3].op));
 	p += 4;
 	break;
+
     default:
-	{
-	    const OP_NAME *q = simple_code;
-	    int k = (p - 1)->op;
-
-	    while (q->op != _HALT && q->op != k)
-		q++;
-
-	    fprintf(fp, "%s\n",
-		    q->op != _HALT ? q->name : "bad instruction");
-	}
+	fprintf(fp, "%s\n", op_name);
 	break;
     }
     return p;
@@ -372,25 +422,37 @@ void
 da(INST * p, FILE *fp)
 {
     INST *base = p;
+
+    TRACE(("dump-asm %s\n",
+	   ((p == begin_start)
+	    ? "BEGIN"
+	    : ((p == end_start)
+	       ? "END"
+	       : ((p == main_start)
+		  ? "MAIN"
+		  : inst2fblock(p)->name)))));
+
     while (p->op != _HALT) {
 	p = da_this(p, base, fp);
     }
     fflush(fp);
 }
 /* *INDENT-OFF* */
-static struct
+static const struct
 {
    PF_CP action ;
-   const char *name ;
+   const char name[10] ;
 }
 special_cases[] =
 {
-   {bi_split,   "split"},
-   {bi_match,   "match"},
-   {bi_getline, "getline"},
-   {bi_sub,     "sub"},
-   {bi_gsub,    "gsub"},
-   {(PF_CP) 0, (char *) 0}
+    { bi_split,    "split" },
+    { bi_length,   "length" },
+    { bi_alength,  "alength" },
+    { bi_match,    "match" },
+    { bi_getline,  "getline" },
+    { bi_sub,      "sub" },
+    { bi_gsub,     "gsub" },
+    { (PF_CP) 0,   "" }
 } ;
 /* *INDENT-ON* */
 
@@ -428,13 +490,14 @@ add_to_fdump_list(FBLOCK * fbp)
 void
 dump_funcs(void)
 {
-    register DUMP_FUNCS *p, *q = fdump_list;
+    DUMP_FUNCS *p;
 
-    while (q) {
-	p = q;
-	q = p->link;
+    TRACE(("dump_funcs\n"));
+
+    while ((p = fdump_list) != NULL) {
 	fprintf(stdout, "function %s\n", p->data->name);
 	da(p->data->code, stdout);
+	fdump_list = p->link;
 	ZFREE(p);
     }
 }
@@ -468,8 +531,6 @@ dump_regex(void)
 {
 }
 #endif /* LOCAL_REGEXP */
-
-#if OPT_TRACE > 0
 /* *INDENT-OFF* */
 static const OP_NAME type_names[] =
 {
@@ -516,6 +577,7 @@ static const OP_NAME other_codes[] = {
     { L_PUSHI,    "l_pushi" },
     { SET_ALOOP,  "set_al" },
     { _CALL,      "call" },
+    { _CALLX,     "callx" },
     { _JMP,       "jmp" },
     { _JNZ,       "jnz" },
     { _JZ,        "jz" },
@@ -530,6 +592,8 @@ static const OP_NAME other_codes[] = {
     { _PUSHINT,   "pushint" },
     { _PUSHS,     "pushs" },
     { _RANGE,     "range" },
+    { _LENGTH,    "defer_len" },
+    { A_LENGTH,   "defer_alen" },
     { _HALT,      "" }
 };
 /* *INDENT-ON* */
@@ -564,9 +628,8 @@ da_op_name(INST * cdp)
 	    }
 	}
 	if (result == 0) {
-	    result = "unknown";
+	    result = "bad instruction";
 	}
     }
     return result;
 }
-#endif
