@@ -1,6 +1,6 @@
 /********************************************
 scan.c
-copyright 2008-2020,2023, Thomas E. Dickey
+copyright 2008-2023,2024, Thomas E. Dickey
 copyright 2010, Jonathan Nieder
 copyright 1991-1996,2014, Michael D. Brennan
 
@@ -12,23 +12,23 @@ the GNU General Public License, version 2, 1991.
 ********************************************/
 
 /*
- * $MawkId: scan.c,v 1.53 2023/08/01 20:50:42 tom Exp $
+ * $MawkId: scan.c,v 1.57 2024/06/22 19:16:39 tom Exp $
  */
 
-#include  "mawk.h"
-#include  "scan.h"
-#include  "memory.h"
-#include  "field.h"
-#include  "init.h"
-#include  "fin.h"
-#include  "repl.h"
-#include  "code.h"
+#include <mawk.h>
+#include <scan.h>
+#include <memory.h>
+#include <field.h>
+#include <init.h>
+#include <fin.h>
+#include <repl.h>
+#include <code.h>
 
-#ifdef	  HAVE_FCNTL_H
-#include  <fcntl.h>
+#ifdef HAVE_FCNTL_H
+#include <fcntl.h>
 #endif
 
-#include  "files.h"
+#include <files.h>
 
 #define CHR_LPAREN '('
 #define CHR_RPAREN ')'
@@ -97,7 +97,7 @@ static int braces_level;
 static int max_repair;
 static struct {
     SYMTAB *stp;
-    char type;
+    SYM_TYPE type;
 } repair_syms[MAX_REPAIR];
 
 /* use unsigned chars for index into scan_code[] */
@@ -213,6 +213,8 @@ scan_fillbuff(void)
 	/* make sure eof is terminated */
 	buffer[r] = '\n';
 	buffer[r + 1] = 0;
+    } else {
+	buffer[r] = 0;
     }
 }
 
